@@ -7,10 +7,12 @@ from collections import namedtuple
 import numpy as np
 
 import unittest
+import warnings
+
 from numba.core.compiler import compile_isolated, Flags
 from numba import jit, typeof, njit, typed
 from numba.core import errors, types, utils, config
-from numba.tests.support import TestCase, tag
+from numba.tests.support import TestCase, tag, ignore_internal_warnings
 
 py38orlater = utils.PYVERSION >= (3, 8)
 
@@ -220,6 +222,112 @@ def sum_kwarg_usecase(x, start=0):
         return ret
 
 
+def isinstance_usecase(a):
+    if isinstance(a, (int, float)):
+        if isinstance(a, int):
+            return a + 1, 'int'
+        if isinstance(a, float):
+            return a + 2.0, 'float'
+    elif isinstance(a, str):
+        return a + ", world!", 'str'
+    elif isinstance(a, complex):
+        return a.imag, 'complex'
+    elif isinstance(a, (tuple, list)):
+        if isinstance(a, tuple):
+            return 'tuple'
+        else:
+            return 'list'
+    elif isinstance(a, set):
+        return 'set'
+    elif isinstance(a, bytes):
+        return 'bytes'
+    return 'no match'
+
+
+def isinstance_dict():
+    a = {1: 2, 3: 4}
+    b = {'a': 10, 'b': np.zeros(3)}
+    if isinstance(a, dict) and isinstance(b, dict):
+        return 'dict'
+    else:
+        return 'not dict'
+
+
+def isinstance_usecase_numba_types(a):
+    if isinstance(a, typed.List):
+        return 'typed list'
+    elif isinstance(a, (types.int32, types.int64)):
+        if isinstance(a, types.int32):
+            return 'int32'
+        else:
+            return 'int64'
+    elif isinstance(a, (types.float32, types.float64)):
+        if isinstance(a, types.float32):
+            return 'float32'
+        elif isinstance(a, types.float64):
+            return 'float64'
+    elif isinstance(a, typed.Dict):
+        return 'typed dict'
+    else:
+        return 'no match'
+
+
+def isinstance_usecase_numba_types_2():
+    # some types cannot be passed as argument to njit functions
+    a = b'hello'
+    b = range(1, 2)
+    c = dict()
+    c[2] = 3
+    if isinstance(a, bytes) and \
+            isinstance(b, range) and \
+            isinstance(c, dict):
+        return True
+    return False
+
+
+def invalid_isinstance_usecase(x):
+    if isinstance(x, ('foo',)):
+        return 'true branch'
+    else:
+        return 'false branch'
+
+
+def isinstance_usecase_invalid_type(x):
+    # this should be a valid call when x := float
+    if isinstance(x, (float, 'not a type')):
+        return True
+    else:
+        return False
+
+
+def invalid_isinstance_usecase_phi_nopropagate(x):
+    if x > 4:
+        z = 10
+    else:
+        z = 'a'
+    if isinstance(z, int):
+        return True
+    else:
+        return False
+
+
+def invalid_isinstance_optional_usecase(x):
+    if x > 4:
+        z = 10
+    else:
+        z = None
+    if isinstance(z, int):
+        return True
+    else:
+        return False
+
+def invalid_isinstance_unsupported_type_usecase():
+    ntpl = namedtuple('ntpl', ['a', 'b'])
+    inst = ntpl(1, 2)
+    def impl(x):
+        return isinstance(inst, ntpl)
+    return impl
+
 class TestBuiltins(TestCase):
 
     def run_nullary_func(self, pyfunc, flags):
@@ -1095,5 +1203,198 @@ class TestOperatorMixedTypes(TestCase):
                 self.assertEqual(expected, got)
 
 
+class TestIsinstanceBuiltin(TestCase):
+    def test_isinstance(self):
+        pyfunc = isinstance_usecase
+        cfunc = jit(nopython=True)(pyfunc)
+
+        inputs = (
+            3,              # int
+            5.0,            # float
+            "Hello",        # string
+            b'world',       # bytes
+            1j,             # complex
+            [1, 2, 3],      # list
+            (1, 3, 3, 3),   # UniTuple
+            set([1, 2]),    # set
+            (1, 'nba', 2),  # Heterogeneous Tuple
+            # {'hello': 2},   # dict - doesn't work as input
+            None,
+        )
+
+        for inpt in inputs:
+            expected = pyfunc(inpt)
+            got = cfunc(inpt)
+            self.assertEqual(expected, got)
+
+    def test_isinstance_dict(self):
+        # Tests typed.Dict and LiteralStrKeyDict
+        pyfunc = isinstance_dict
+        cfunc = jit(nopython=True)(pyfunc)
+        self.assertEqual(pyfunc(), cfunc())
+
+    def test_isinstance_numba_types(self):
+        # This makes use of type aliasing between python scalars and NumPy
+        # scalars, see also test_numba_types()
+        pyfunc = isinstance_usecase_numba_types
+        cfunc = jit(nopython=True)(pyfunc)
+
+        inputs = (
+            (types.int32(1), 'int32'),
+            (types.int64(2), 'int64'),
+            (types.float32(3.0), 'float32'),
+            (types.float64(4.0), 'float64'),
+            (types.complex64(5j), 'no match'),
+            (typed.List([1, 2]), 'typed list'),
+            (typed.Dict.empty(types.int64, types.int64), 'typed dict')
+        )
+
+        for inpt, expected in inputs:
+            got = cfunc(inpt)
+            self.assertEqual(expected, got)
+
+    def test_isinstance_numba_types_2(self):
+        pyfunc = isinstance_usecase_numba_types_2
+        cfunc = jit(nopython=True)(pyfunc)
+        self.assertEqual(pyfunc(), cfunc())
+
+    def test_isinstance_invalid_type(self):
+        pyfunc = isinstance_usecase_invalid_type
+        cfunc = jit(nopython=True)(pyfunc)
+
+        # valid type
+        self.assertTrue(cfunc(3.4))
+
+        # invalid type
+        msg = 'Cannot infer numba type of python type'
+
+        with self.assertRaises(errors.TypingError) as raises:
+            cfunc(100)
+
+        self.assertIn(msg, str(raises.exception))
+
+    def test_isinstance_exceptions(self):
+        fns = [
+            (invalid_isinstance_usecase,
+             'Cannot infer numba type of python type'),
+            (invalid_isinstance_usecase_phi_nopropagate,
+             ('isinstance() cannot determine the type of variable "z" due to a '
+             'branch.')),
+            (invalid_isinstance_optional_usecase,
+             ('isinstance() cannot determine the type of variable "z" due to a '
+             'branch.')),
+            (invalid_isinstance_unsupported_type_usecase(),
+             ('isinstance() does not support variables of type "ntpl(')),
+        ]
+
+        for fn, msg in fns:
+            fn = njit(fn)
+
+            with self.assertRaises(errors.TypingError) as raises:
+                fn(100)
+
+            self.assertIn(msg, str(raises.exception))
+
+    def test_combinations(self):
+        # Combinatorically test common classes and instances
+        def gen_w_arg(clazz_type):
+            def impl(x):
+                return isinstance(x, clazz_type)
+            return impl
+
+        clazz_types = (int, float, complex, str, list, tuple, bytes, set, range,
+                       np.int8, np.float32,)
+        instances = (1, 2.3, 4j, '5', [6,], (7,), b'8', {9,}, None,
+                     (10, 11, 12), (13, 'a', 14j), np.array([15, 16, 17]),
+                     np.int8(18), np.float32(19),
+                     typed.Dict.empty(types.unicode_type, types.float64),
+                     typed.List.empty_list(types.complex128), np.ones(4))
+
+        for ct in clazz_types:
+            fn = njit(gen_w_arg(ct))
+            for x in instances:
+                expected = fn.py_func(x)
+                got = fn(x)
+                self.assertEqual(got, expected)
+
+    def test_numba_types(self):
+        # Check types which are Numba types, this would break without the jit
+        # decorator in all cases except numba.typed containers.
+        def gen_w_arg(clazz_type):
+            def impl():
+                return isinstance(1, clazz_type)
+            return impl
+
+        clazz_types = (types.Integer, types.Float, types.Array,)
+
+        msg = "Numba type classes.*are not supported"
+        for ct in clazz_types:
+            fn = njit(gen_w_arg(ct))
+            with self.assertRaises(errors.TypingError) as raises:
+                fn()
+            self.assertRegex(str(raises.exception), msg)
+
+    def test_python_numpy_scalar_alias_problem(self):
+        # There's a problem due to Python and NumPy scalars being aliased in the
+        # type system. This is because e.g. int scalar values and NumPy np.intp
+        # type alias to types.intp. This test merely records this fact.
+
+        @njit
+        def foo():
+            return isinstance(np.intp(10), int)
+
+        self.assertEqual(foo(), True)
+        self.assertEqual(foo.py_func(), False)
+
+        @njit
+        def bar():
+            return isinstance(1, np.intp)
+
+        self.assertEqual(bar(), True)
+        self.assertEqual(bar.py_func(), False)
+
+    def test_branch_prune(self):
+        # Check that isinstance branches are pruned allowing otherwise
+        # impossible type specific specialisation.
+
+        @njit
+        def foo(x):
+            if isinstance(x, str):
+                return x + 'some_string'
+            elif isinstance(x, complex):
+                return np.imag(x)
+            elif isinstance(x, tuple):
+                return len(x)
+            else:
+                assert 0
+
+        for x in ('string', 1 + 2j, ('a', 3, 4j)):
+            expected = foo.py_func(x)
+            got = foo(x)
+            self.assertEqual(got, expected)
+
+    def test_experimental_warning(self):
+        # Check that if the isinstance feature is in use then an experiemental
+        # warning is raised.
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter('always', errors.NumbaWarning)
+            ignore_internal_warnings()
+
+            @njit
+            def foo(x):
+                return isinstance(x, float)
+
+            foo(1.234)
+
+            self.assertEqual(len(w), 1)
+
+            self.assertEqual(w[0].category,
+                             errors.NumbaExperimentalFeatureWarning)
+            msg = ("Use of isinstance() detected. This is an experimental "
+                   "feature.")
+            self.assertIn(msg, str(w[0].message))
+
+
 if __name__ == '__main__':
     unittest.main()
